from datetime import datetime
from pypomes_core import exc_format
from typing import Final
import logging
import pytz
import re
import sys
from .threaded_scheduler import __ThreadedScheduler

__REGEX_VERIFY_CRON: Final[str] = "/(@(annually|yearly|monthly|weekly|daily|hourly|reboot))|" \
                                  "(@every (\d+(ns|us|µs|ms|s|m|h))+)|((((\d+,)+\d+|(\d+(\/|-)\d+)|\d+|\*) ?){5,7})"

__scheduler: __ThreadedScheduler


def scheduler_create(errors: list[str], timezone: pytz.BaseTzInfo,
                     retry_interval: int, logger: logging.Logger = None) -> bool:
    """
    Create the threaded job scheduler. This is a wrapper around the package *apscheduler*.

    :param errors: Resulting errors, if any
    :param timezone: The timezone to be used
    :param retry_interval: Interval between retry attempts, in minutes
    :param logger: Optional logger object
    :return: True if scheduler was created, or False if an error prevented its creation
    """
    global __scheduler
    # inicialize the return variable
    result: bool = False

    try:
        __scheduler = __ThreadedScheduler(timezone, retry_interval, logger)
        __scheduler.daemon = True
        result = True
    except Exception as e:
        errors.append(f"Error creating the job scheduler: {exc_format(e, sys.exc_info())}")

    return result


def scheduler_schedule(errors: list[str],
                       jobs: list[tuple[callable, str, str, str, datetime, tuple, dict]]) -> int:
    """
    Schedule the jobs in *jobs*, starting at the given *start*. Each element in the job list
    is a *tuple* with the corresponding job data:
    *(callable function, job id, job name, CRON expression, job start timestamp, job args, job kwargs)*.

    :param errors: Resulting errors, if any
    :param jobs: List of tuples with jobs to schedule
    :return: The number of jobs effectively scheduled
    """
    global __scheduler
    # inicializa a variável de retorno
    result: int = 0

    # traverse the job list and do the scheduling
    for job in jobs:

        # is it a valid CRON expression ?
        if re.search(__REGEX_VERIFY_CRON, job[3]) is None:
            # no, report the error
            errors.append(f"Invalid CRON expression: '{job[4]}'")
        else:
            # yes, proceed with the scheduling
            try:
                __scheduler.schedule_job(job[0], job[1], job[2], job[3], job[4], job[5], job[6])
                result += 1
            except Exception as e:
                errors.append(f"Error scheduling the job '{job[2]}', id '{job[1]}', "
                              f"with CRON '{job[3]}': {exc_format(e, sys.exc_info())}")

    return result


def scheduler_start():
    """
    Start the scheduler.
    """
    global __scheduler
    __scheduler.start()


def scheduler_stop():
    """
    Stop the scheduler.
    """
    global __scheduler
    __scheduler.stop()



from datetime import datetime
from apscheduler.schedulers.blocking import BlockingScheduler
from apscheduler.triggers.cron import CronTrigger
import logging
import pytz
import threading


class __ThreadedScheduler(threading.Thread):
    """
    A scalable implementation of *APScheduler*'s *BlockingScheduler*, running as single or multiple instances,
    each on its own thread.
    """
    _scheduler: BlockingScheduler
    _logger: logging.Logger
    _stopped: bool

    def __init__(self, timezone: pytz.timezone, retry_interval: int, logger: logging.Logger = None):

        threading.Thread.__init__(self)

        self._stopped = False
        self._logger = logger
        self._scheduler = BlockingScheduler(logging=logger,
                                            timezone=timezone,
                                            jobstore_retry_interval=retry_interval)
        if self._logger is not None:
            self._logger.info("Instanced, with timezone "
                              f"'{timezone}' and retry interval '{retry_interval}'")

    def run(self):
        """
        Start the scheduler in itas own thread.
        """
        # stay in loop until 'stop()' is invoked
        while not self._stopped:
            if self._logger is not None:
                self._logger.info("Started")

            # start the scheduler, blocking until it is interrupted
            self._scheduler.start()

        self._scheduler.shutdown()
        if self._logger is not None:
            self._logger.info("Finished")

    def schedule_job(self, job: callable, job_id: str, job_name: str, job_cron: str,
                     job_start: datetime = None, job_args: tuple = None, job_kwargs: dict = None):
        """
        Schedule the given *job*, with the parameters
        :param job: the callable object to be scheduled
        :param job_id: the id of the scheduled job
        :param job_name: the name of the scheduled job
        :param job_cron: the CRON expression directing the execution times
        :param job_start: the start timestamp for the scheduling process
        :param job_args: the '*args' arguments to be passed to the scheduled job
        :param job_kwargs: the '**kwargs' arguments to be passed to the scheduled job
        :return:
        """
        # approximately, convert symbols to CRON expression
        expr: str
        match job_cron:
            case "@reboot":
                expr = "1 0 * * *"                      # daily, at 00h01
            case "@midnight":
                expr = "0 0 * * *"                      # daily, at 00h00
            case "@hourly":
                expr = "0 * * * *"                      # every hour, at minute 0 (??h00)
            case "@daily":
                expr = "1 0 * * *"                      # daily, at 00h01
            case "@weekly":
                expr = "0 0 * * 0"                      # on sundays, at 00h00
            case "@monthly":
                expr = "0 0 1 * *"                      # on the first day of the month, at 00h00
            case "@yearly" | "@annually":
                expr = "0 0 1 1 *"                      # on January 1st, at 00h00
            case _:
                expr = job_cron

        # CRON expression: <minute> <hour> <day-of-month> <month> <day-of-week>
        vals: list[str] = expr.split()
        vals = [None if val == '?' else val for val in vals]
        aps_trigger = CronTrigger(minute=vals[0],
                                  hour=vals[1],
                                  day=vals[2],
                                  month=vals[3],
                                  day_of_week=vals[4],
                                  start_date=job_start)
        self._scheduler.add_job(func=job,
                                trigger=aps_trigger,
                                args=job_args,
                                kwargs=job_kwargs,
                                id=job_id,
                                name=job_name)
        if self._logger is not None:
            self._logger.info(f"Job '{job_name}' scheduled, with CRON '{job_cron}'")

    def stop(self):
        """
        Stop the scheduler.
        :return:
        """
        if self._logger is not None:
            self._logger.info("Finishing...")
        self._stopped = True



from .scheduling_pomes import (
    scheduler_create, scheduler_schedule, scheduler_start, scheduler_stop
)

__all__ = [
    # scheduling_pomes
    scheduler_create, scheduler_schedule, scheduler_start, scheduler_stop
]

__version__ = "0.1.6"
__version_info__ = (0, 1, 6)
